import tensorflow as tf

from stable_baselines.common.policies import ActorCriticPolicy
from stable_baselines.common.tf_layers import conv, linear, conv_to_fc

_Z = 32
_RS = 5
_64 = 64
_C = 64  # controller FC layer size
ARCH = "VCARCH"

# Custom MLP policy of three layers of size 128 each for the actor and 2 layers of 32 for the critic,
# with a nature_cnn feature extractor
class Custom1DPolicy(ActorCriticPolicy):
    def __init__(self, sess, ob_space, ac_space, n_env, n_steps, n_batch, reuse=False, **kwargs):
        super(Custom1DPolicy, self).__init__(
            sess, ob_space, ac_space, n_env, n_steps, n_batch, reuse=reuse, scale=True)

        with tf.variable_scope("model", reuse=reuse):
            activ = tf.nn.relu

            x = self.processed_obs[:, :-_RS]
            state_features = tf.reshape(self.processed_obs[:, -_RS:], (-1, _RS))

            h = tf.layers.conv1d(
                x, 32, 8, strides=4, activation=tf.nn.relu, name="enc_conv1"
            )
            h = tf.layers.conv1d(
                h, 64, 9, strides=4, activation=tf.nn.relu, name="enc_conv2"
            )
            h = tf.layers.conv1d(
                h, 128, 6, strides=4, activation=tf.nn.relu, name="enc_conv3"
            )
            h = tf.layers.conv1d(
                h, 256, 4, strides=4, activation=tf.nn.relu, name="enc_conv4"
            )
            h = tf.reshape(h, [-1, 4 * 256])
            h = tf.layers.dense(h, _Z, name="enc_fc_mu")
            extracted_features = tf.concat([h, state_features], axis=1)

            pi_h = extracted_features
            for i, layer_size in enumerate([_C, _C]):
                pi_h = activ(tf.layers.dense(pi_h, layer_size, name='pi_fc' + str(i)))
            pi_latent = pi_h

            vf_h = extracted_features
            for i, layer_size in enumerate([_C, _C]):
                vf_h = activ(tf.layers.dense(vf_h, layer_size, name='vf_fc' + str(i)))
            value_fn = tf.layers.dense(vf_h, 1, name='vf')
            vf_latent = vf_h

            self._proba_distribution, self._policy, self.q_value = \
                self.pdtype.proba_distribution_from_latent(pi_latent, vf_latent, init_scale=0.01)

        self._value_fn = value_fn
        self._setup_init()

    def step(self, obs, state=None, mask=None, deterministic=False):
        if deterministic:
            action, value, neglogp = self.sess.run([self.deterministic_action, self.value_flat, self.neglogp],
                                                   {self.obs_ph: obs})
        else:
            action, value, neglogp = self.sess.run([self.action, self.value_flat, self.neglogp],
                                                   {self.obs_ph: obs})
        return action, value, self.initial_state, neglogp

    def proba_step(self, obs, state=None, mask=None):
        return self.sess.run(self.policy_proba, {self.obs_ph: obs})

    def value(self, obs, state=None, mask=None):
        return self.sess.run(self.value_flat, {self.obs_ph: obs})

class CustomPolicy(ActorCriticPolicy):
    def __init__(self, sess, ob_space, ac_space, n_env, n_steps, n_batch, reuse=False, **kwargs):
        super(CustomPolicy, self).__init__(
            sess, ob_space, ac_space, n_env, n_steps, n_batch, reuse=reuse, scale=True)

        with tf.variable_scope("model", reuse=reuse):
            activ = tf.nn.relu

            x = self.processed_obs[:, :-_RS]
            x = tf.reshape(x, (-1, _64, _64, 1))
            state_features = tf.reshape(self.processed_obs[:, -_RS:], (-1, _RS))

            # Encoder
            h = tf.layers.conv2d(
                x, 32, 4, strides=2, activation=tf.nn.relu, name="enc_conv1"
            )
            h = tf.layers.conv2d(
                h, 64, 4, strides=2, activation=tf.nn.relu, name="enc_conv2"
            )
            h = tf.layers.conv2d(
                h, 128, 4, strides=2, activation=tf.nn.relu, name="enc_conv3"
            )
            h = tf.layers.conv2d(
                h, 256, 4, strides=2, activation=tf.nn.relu, name="enc_conv4"
            )
            h = tf.reshape(h, [-1, 2 * 2 * 256])
            h = tf.layers.dense(h, _Z, name="enc_fc_mu")
            extracted_features = tf.concat([h, state_features], axis=1)

            pi_h = extracted_features
            for i, layer_size in enumerate([_C, _C]):
                pi_h = activ(tf.layers.dense(pi_h, layer_size, name='pi_fc' + str(i)))
            pi_latent = pi_h

            vf_h = extracted_features
            for i, layer_size in enumerate([_C, _C]):
                vf_h = activ(tf.layers.dense(vf_h, layer_size, name='vf_fc' + str(i)))
            value_fn = tf.layers.dense(vf_h, 1, name='vf')
            vf_latent = vf_h

            self._proba_distribution, self._policy, self.q_value = \
                self.pdtype.proba_distribution_from_latent(pi_latent, vf_latent, init_scale=0.01)

        self._value_fn = value_fn
        self._setup_init()

    def step(self, obs, state=None, mask=None, deterministic=False):
        if deterministic:
            action, value, neglogp = self.sess.run([self.deterministic_action, self.value_flat, self.neglogp],
                                                   {self.obs_ph: obs})
        else:
            action, value, neglogp = self.sess.run([self.action, self.value_flat, self.neglogp],
                                                   {self.obs_ph: obs})
        return action, value, self.initial_state, neglogp

    def proba_step(self, obs, state=None, mask=None):
        return self.sess.run(self.policy_proba, {self.obs_ph: obs})

    def value(self, obs, state=None, mask=None):
        return self.sess.run(self.value_flat, {self.obs_ph: obs})



from __future__ import print_function
import numpy as np
import os
import pandas as pd
from pyniel.python_tools.path_tools import make_dir_if_not_exists
import time
from datetime import datetime

from navrep.tools.data_extraction import archive_to_lidar_dataset
from navrep.tools.rings import generate_rings
from navrep.models.vae2d import ConvVAE, reset_graph
from navrep.tools.commonargs import parse_common_args

_Z = 32

if __name__ == "__main__":
    common_args, _ = parse_common_args()
    VARIANT = common_args.environment
    print(common_args)

    # Parameters for training
    batch_size = 100
    DATA_DIR = "record"
    START_TIME = datetime.now().strftime("%Y_%m_%d__%H_%M_%S")
    MAX_STEPS = common_args.n
    if MAX_STEPS is None:
        MAX_STEPS = 222222
    NUM_EPOCH = MAX_STEPS  # don't limit based on epoch

    if VARIANT == "ian":
        model_save_path = os.path.expanduser("~/navrep/models/V/vae.json")
        dataset_dir = os.path.expanduser("~/navrep/datasets/V/ian")
        log_path = os.path.expanduser("~/navrep/logs/V/vae_train_log_{}.csv".format(START_TIME))
    if VARIANT == "toy":
        model_save_path = os.path.expanduser("~/navrep/models/V/toyvae.json")
        dataset_dir = os.path.expanduser("~/navrep/datasets/V/toy")
        log_path = os.path.expanduser("~/navrep/logs/V/toyvae_train_log_{}.csv".format(START_TIME))
    if VARIANT == "markone":
        model_save_path = os.path.expanduser("~/navrep/models/V/markonevae.json")
        dataset_dir = os.path.expanduser("~/navrep/datasets/V/markone")
        log_path = os.path.expanduser("~/navrep/logs/V/markonevae_train_log_{}.csv".format(START_TIME))
    if VARIANT == "marktwo":
        model_save_path = os.path.expanduser("~/navrep/models/V/marktwovae.json")
        dataset_dir = os.path.expanduser("~/navrep/datasets/V/marktwo")
        log_path = os.path.expanduser("~/navrep/logs/V/marktwovae_train_log_{}.csv".format(START_TIME))
    if VARIANT == "navreptrain":
        model_save_path = os.path.expanduser("~/navrep/models/V/navreptrainvae.json")
        dataset_dir = os.path.expanduser("~/navrep/datasets/V/navreptrain")
        log_path = os.path.expanduser("~/navrep/logs/V/navreptrainvae_train_log_{}.csv".format(START_TIME))
    if VARIANT == "irl":
        model_save_path = os.path.expanduser("~/navrep/models/V/irlvae.json")
        dataset_dir = os.path.expanduser("~/navrep/datasets/V/irl")
        log_path = os.path.expanduser("~/navrep/logs/V/irlvae_train_log_{}.csv".format(START_TIME))

    if common_args.dry_run:
        model_save_path = model_save_path.replace(os.path.expanduser("~/navrep"), "/tmp/navrep")
        log_path = log_path.replace(os.path.expanduser("~/navrep"), "/tmp/navrep")

    make_dir_if_not_exists(os.path.dirname(model_save_path))
    make_dir_if_not_exists(os.path.dirname(log_path))

    # create network
    reset_graph()
    vae = ConvVAE(z_size=_Z, batch_size=batch_size, is_training=True, reuse=False)
    vae.print_trainable_params()

    # create training dataset
    dataset = archive_to_lidar_dataset(dataset_dir)
    if len(dataset) == 0:
        raise ValueError("no scans found, exiting")
    print(len(dataset), "scans in dataset.")

    # split into batches:
    total_length = len(dataset)
    num_batches = int(np.floor(total_length / batch_size))

    # rings converter
    rings_def = generate_rings(64, 64)

    # train loop:
    print("train", "step", "loss", "recon_loss", "kl_loss")
    values_logs = None
    start = time.time()
    for epoch in range(NUM_EPOCH):
        np.random.shuffle(dataset)
        for idx in range(num_batches):
            batch = dataset[idx * batch_size : (idx + 1) * batch_size]
            scans = batch

            obs = (
                rings_def["lidar_to_rings"](scans).astype(float)
                / rings_def["rings_to_bool"]
            )
            # remove "too close" points
            obs[:, :, 0, :] = 0.0

            feed = {
                vae.x: obs,
            }

            (train_loss, r_loss, kl_loss, train_step, _) = vae.sess.run(
                [vae.loss, vae.r_loss, vae.kl_loss, vae.global_step, vae.train_op], feed
            )

            if (train_step + 1) % 500 == 0:
                if common_args.render:
                    from matplotlib import pyplot as plt

                    #           plt.ion()
                    plt.figure("training_status")
                    plt.clf()
                    plt.suptitle("training step {}".format(train_step))
                    f, (ax1, ax2) = plt.subplots(2, 1, num="training_status")
                    ax1.imshow(obs[0, :, :, 0], cmap=plt.cm.Greys)
                    ax2.imshow(vae.encode_decode(obs)[0, :, :, 0], cmap=plt.cm.Greys)
                    plt.savefig("/tmp/vae_step{:07}.png".format(train_step))
                #           plt.pause(0.01)
                print("step", (train_step + 1), train_loss, r_loss, kl_loss)
                vae.save_json(model_save_path)
                # log
                end = time.time()
                time_taken = end - start
                start = time.time()
                values_log = pd.DataFrame(
                    [[train_step, train_loss, r_loss, kl_loss, time_taken]],
                    columns=["step", "cost", "rec_cost", "kl_cost", "train_time_taken"],
                )
                if values_logs is None:
                    values_logs = values_log.copy()
                else:
                    values_logs = values_logs.append(values_log, ignore_index=True)
                values_logs.to_csv(log_path)

            if MAX_STEPS is not None:
                if train_step > MAX_STEPS:
                    break
        if MAX_STEPS is not None:
            if train_step > MAX_STEPS:
                break

    # finished, final model:
    vae.save_json(model_save_path)



from __future__ import print_function
import numpy as np
import os
import tensorflow as tf
from collections import namedtuple
import json
import pickle

MAX_GOAL_DIST = 25.  # for normalization
_Z = 32  # V latent shape
_H = 512
_G = 2  # robot states (goal)
_A = 3  # action size

# controls whether we concatenate (z, c, h), etc for features used for car.
MODE_ZCH = 0
MODE_ZC = 1
MODE_Z = 2
MODE_Z_HIDDEN = 3  # extra hidden later
MODE_ZH = 4

HyperParams = namedtuple(
    "HyperParams",
    [
        "max_seq_len",
        "seq_width",
        "rnn_size",
        "action_width",
        "batch_size",
        "grad_clip",
        "num_mixture",
        "restart_factor",
        "learning_rate",
        "decay_rate",
        "min_learning_rate",
        "use_layer_norm",
        "use_recurrent_dropout",
        "recurrent_dropout_prob",
        "use_input_dropout",
        "input_dropout_prob",
        "use_output_dropout",
        "output_dropout_prob",
        "is_training",
        "differential_z",
        "residual_z",
    ],
)


def default_hps():
    return HyperParams(
        max_seq_len=1000,  # train on sequences of 1000
        seq_width=_Z+_G,  # width of our data (32 + 2)
        rnn_size=_H,  # number of rnn cells
        action_width=_A,
        batch_size=32,  # minibatch sizes
        grad_clip=1.0,
        num_mixture=5,  # number of mixtures in MDN
        restart_factor=10.0,  # factor of importance for restart=1 rare case for loss.
        learning_rate=0.001,
        decay_rate=0.99999,
        min_learning_rate=0.00001,
        use_layer_norm=0,  # set this to 1 to get more stable results (less chance of NaN), but slower
        use_recurrent_dropout=0,
        recurrent_dropout_prob=0.90,
        use_input_dropout=0,
        input_dropout_prob=0.90,
        use_output_dropout=0,
        output_dropout_prob=0.90,
        is_training=1,
        differential_z=True,
        residual_z=False,
    )


default_hps_params = default_hps()
sample_hps_params = default_hps_params._replace(
    batch_size=1, max_seq_len=2, use_recurrent_dropout=0, is_training=0
)


def reset_graph():
    if "sess" in globals() and sess:  # noqa
        sess.close()  # noqa
    tf.reset_default_graph()


# MDN-RNN model tailored for doomrnn
class MDNRNN:
    def __init__(self, hps, gpu_mode=True, reuse=False):
        self.hps = hps
        with tf.variable_scope("MDNRNN", reuse=reuse):
            if not gpu_mode:
                with tf.device("/cpu:0"):
                    print("model using cpu")
                    self.g = tf.Graph()
                    with self.g.as_default():
                        self.build_model(hps)
            else:
                print("model using gpu")
                self.g = tf.Graph()
                with self.g.as_default():
                    self.build_model(hps)
        self.init_session()

    def build_model(self, hps):

        self.num_mixture = hps.num_mixture
        KMIX = self.num_mixture  # 5 mixtures
        ACTION_WIDTH = hps.action_width
        LENGTH = self.hps.max_seq_len - 1  # 999 timesteps

        if hps.is_training:
            self.global_step = tf.Variable(0, name="global_step", trainable=False)

        cell_fn = tf.contrib.rnn.LayerNormBasicLSTMCell  # use LayerNormLSTM

        use_recurrent_dropout = False if self.hps.use_recurrent_dropout == 0 else True
        use_input_dropout = False if self.hps.use_input_dropout == 0 else True
        use_output_dropout = False if self.hps.use_output_dropout == 0 else True
        use_layer_norm = False if self.hps.use_layer_norm == 0 else True

        if use_recurrent_dropout:
            cell = cell_fn(
                hps.rnn_size,
                layer_norm=use_layer_norm,
                dropout_keep_prob=self.hps.recurrent_dropout_prob,
            )
        else:
            cell = cell_fn(hps.rnn_size, layer_norm=use_layer_norm)

        # multi-layer, and dropout:
        print("input dropout mode =", use_input_dropout)
        print("output dropout mode =", use_output_dropout)
        print("recurrent dropout mode =", use_recurrent_dropout)
        if use_input_dropout:
            print(
                "applying dropout to input with keep_prob =",
                self.hps.input_dropout_prob,
            )
            cell = tf.contrib.rnn.DropoutWrapper(
                cell, input_keep_prob=self.hps.input_dropout_prob
            )
        if use_output_dropout:
            print(
                "applying dropout to output with keep_prob =",
                self.hps.output_dropout_prob,
            )
            cell = tf.contrib.rnn.DropoutWrapper(
                cell, output_keep_prob=self.hps.output_dropout_prob
            )
        self.cell = cell

        self.sequence_lengths = LENGTH  # assume every sample has same length.
        self.batch_z_rs = tf.placeholder(  # z (34) + robotstate (2)
            dtype=tf.float32,
            shape=[self.hps.batch_size, self.hps.max_seq_len, hps.seq_width],
        )
        self.batch_action = tf.placeholder(
            dtype=tf.float32,
            shape=[self.hps.batch_size, self.hps.max_seq_len, ACTION_WIDTH],
        )
        self.batch_restart = tf.placeholder(
            dtype=tf.float32, shape=[self.hps.batch_size, self.hps.max_seq_len]
        )
        # TODO: add latent goal info

        self.input_z = self.batch_z_rs[:, :LENGTH, :]
        self.input_action = self.batch_action[:, :LENGTH, :]
        self.input_restart = self.batch_restart[:, :LENGTH]

        if hps.differential_z:
            self.target_z = (
                self.batch_z_rs[:, 1:, :] - self.batch_z_rs[:, :LENGTH, :]
            )  # delta Z prediction
        else:
            self.target_z = self.batch_z_rs[:, 1:, :]

        self.target_restart = self.batch_restart[:, 1:]

        self.input_seq = tf.concat(
            [
                self.input_z,
                self.input_action,
                tf.reshape(self.input_restart, [self.hps.batch_size, LENGTH, 1]),
            ],
            axis=2,
        )

        self.zero_state = cell.zero_state(batch_size=hps.batch_size, dtype=tf.float32)
        self.initial_state = self.zero_state

        inputs = tf.unstack(self.input_seq, axis=1)

        def custom_rnn_autodecoder(
            decoder_inputs, input_restart, initial_state, cell, scope=None
        ):
            # customized rnn_decoder for the task of dealing with restart
            with tf.variable_scope(scope or "MDNRNN"):
                state = initial_state
                zero_c, zero_h = self.zero_state
                outputs = []

                for i in range(LENGTH):
                    inp = decoder_inputs[i]
                    if i > 0:
                        tf.get_variable_scope().reuse_variables()

                    # if restart is 1, then set lstm state to zero
                    restart_flag = tf.greater(input_restart[:, i], 0.5)

                    c, h = state

                    c = tf.where(restart_flag, zero_c, c)
                    h = tf.where(restart_flag, zero_h, h)

                    output, state = cell(inp, tf.nn.rnn_cell.LSTMStateTuple(c, h))
                    outputs.append(output)

            return outputs, state

        outputs, final_state = custom_rnn_autodecoder(
            inputs, self.input_restart, self.initial_state, self.cell
        )
        output = tf.reshape(tf.concat(outputs, axis=1), [-1, self.hps.rnn_size])

        NOUT = hps.seq_width * KMIX * 3 + 1  # plus 1 to predict the restart state.

        with tf.variable_scope("MDNRNN"):
            output_w = tf.get_variable("output_w", [self.hps.rnn_size, NOUT])
            output_b = tf.get_variable("output_b", [NOUT])

        output = tf.reshape(output, [-1, hps.rnn_size])  # (batch*seq, 512)
        output = tf.nn.xw_plus_b(output, output_w, output_b)  # (batch*seq, NOUT)

        self.out_restart_logits = output[:, 0]
        output = output[:, 1:]  # (batch*seq, 34 * 5 * 3)

        output = tf.reshape(output, [-1, KMIX * 3])  # (batch*seq*34, 5*3)
        self.final_state = final_state

        logSqrtTwoPI = np.log(np.sqrt(2.0 * np.pi))

        def tf_lognormal(y, mean, logstd):
            return -0.5 * ((y - mean) / tf.exp(logstd)) ** 2 - logstd - logSqrtTwoPI

        def get_lossfunc(logmix, mean, logstd, y):
            v = logmix + tf_lognormal(y, mean, logstd)
            v = tf.reduce_logsumexp(v, 1, keepdims=True)
            return -tf.reduce_mean(v)

        def get_mdn_coef(output):
            logmix, mean, logstd = tf.split(output, 3, 1)
            logmix = logmix - tf.reduce_logsumexp(logmix, 1, keepdims=True)
            return logmix, mean, logstd

        out_logmix, out_mean, out_logstd = get_mdn_coef(output)

        if hps.residual_z:
            flat_input_data = tf.reshape(self.input_z, [-1, 1])  # (batch*seq*34, 1)
            out_mean = out_mean + flat_input_data

        self.out_logmix = out_logmix  # (batch*seq*34, 5)
        self.out_mean = out_mean
        self.out_logstd = out_logstd

        # reshape target data so that it is compatible with prediction shape
        flat_target_data = tf.reshape(self.target_z, [-1, 1])

        lossfunc = get_lossfunc(out_logmix, out_mean, out_logstd, flat_target_data)

        self.z_cost = tf.reduce_mean(lossfunc)

        flat_target_restart = tf.reshape(self.target_restart, [-1, 1])

        self.r_cost = tf.nn.sigmoid_cross_entropy_with_logits(
            labels=flat_target_restart,
            logits=tf.reshape(self.out_restart_logits, [-1, 1]),
        )

        factor = tf.ones_like(self.r_cost) + flat_target_restart * (
            self.hps.restart_factor - 1.0
        )

        self.r_cost = tf.reduce_mean(tf.multiply(factor, self.r_cost))

        self.cost = self.z_cost + self.r_cost

        if self.hps.is_training == 1:
            self.lr = tf.Variable(self.hps.learning_rate, trainable=False)
            optimizer = tf.train.AdamOptimizer(self.lr)

            gvs = optimizer.compute_gradients(self.cost)
            capped_gvs = [
                (tf.clip_by_value(grad, -self.hps.grad_clip, self.hps.grad_clip), var)
                for grad, var in gvs
            ]
            self.train_op = optimizer.apply_gradients(
                capped_gvs, global_step=self.global_step, name="train_step"
            )

        # initialize vars
        self.init = tf.global_variables_initializer()

        t_vars = tf.trainable_variables()
        self.assign_ops = {}
        for var in t_vars:
            if var.name.startswith("MDNRNN"):
                pshape = var.get_shape()
                pl = tf.placeholder(tf.float32, pshape, var.name[:-2] + "_placeholder")
                assign_op = var.assign(pl)
                self.assign_ops[var] = (assign_op, pl)

    def init_session(self):
        """Launch TensorFlow session and initialize variables"""
        self.sess = tf.Session(graph=self.g)
        self.sess.run(self.init)

    def close_sess(self):
        """ Close TensorFlow session """
        self.sess.close()

    def save_model(self, model_save_path, epoch):
        sess = self.sess
        with self.g.as_default():
            saver = tf.train.Saver(tf.global_variables())
        checkpoint_path = os.path.join(model_save_path, "doomcover_rnn")
        tf.logging.info("saving model %s.", checkpoint_path)
        saver.save(sess, checkpoint_path, epoch)  # just keep one

    def load_checkpoint(self, checkpoint_path):
        sess = self.sess
        with self.g.as_default():
            saver = tf.train.Saver(tf.global_variables())
        ckpt = tf.train.get_checkpoint_state(checkpoint_path)
        print("loading model", ckpt.model_checkpoint_path)
        tf.logging.info("Loading model %s.", ckpt.model_checkpoint_path)
        saver.restore(sess, ckpt.model_checkpoint_path)

    def get_model_params(self):
        # get trainable params.
        model_names = []
        model_params = []
        model_shapes = []
        with self.g.as_default():
            t_vars = tf.trainable_variables()
            for var in t_vars:
                if var.name.startswith("MDNRNN"):
                    param_name = var.name
                    p = self.sess.run(var)
                    model_names.append(param_name)
                    params = np.round(p * 10000).astype(np.int).tolist()
                    model_params.append(params)
                    model_shapes.append(p.shape)
        return model_params, model_shapes, model_names

    def print_trainable_params(self):
        trainable_params = 0
        with self.g.as_default():
            t_vars = tf.trainable_variables()
            for var in t_vars:
                if var.name.startswith("MDNRNN"):
                    trainable_params += np.prod(var.get_shape().as_list())
        print("RNN Trainable parameters: {}".format(trainable_params))

    def set_model_params(self, params):
        with self.g.as_default():
            t_vars = tf.trainable_variables()
            idx = 0
            for var in t_vars:
                if var.name.startswith("MDNRNN"):
                    pshape = tuple(var.get_shape().as_list())
                    p = np.array(params[idx])
                    assert pshape == p.shape, "inconsistent shape {}, {}".format(pshape, p.shape)
                    assign_op, pl = self.assign_ops[var]
                    self.sess.run(assign_op, feed_dict={pl.name: p / 10000.0})
                    idx += 1

    def get_random_model_params(self, stdev=0.5):
        # get random params.
        _, mshape, _ = self.get_model_params()
        rparam = []
        for s in mshape:
            # rparam.append(np.random.randn(*s)*stdev)
            rparam.append(np.random.standard_cauchy(s) * stdev)  # spice things up!
        return rparam

    def set_random_params(self, stdev=0.5):
        rparam = self.get_random_model_params(stdev)
        self.set_model_params(rparam)

    def load_json(self, jsonfile="rnn.json"):
        with open(jsonfile, "r") as f:
            params = json.load(f)
        self.set_model_params(params)

    def save_json(self, jsonfile="rnn.json"):
        model_params, model_shapes, model_names = self.get_model_params()
        qparams = []
        for p in model_params:
            qparams.append(p)
        with open(jsonfile, "wt") as outfile:
            json.dump(
                qparams, outfile, sort_keys=True, indent=0, separators=(",", ": ")
            )
        base = jsonfile
        if jsonfile.endswith(".json"):
            base = os.path.splitext(jsonfile)[0]
        with open(base+".hyperparams.pckl", "wb") as f:
            pickle.dump(self.hps, f)
        print("{} written.".format(jsonfile))

    def rnn_output_size(self, mode):
        if mode == MODE_ZCH:
            return self.hps.seq_width + self.hps.rnn_size + self.hps.rnn_size
        if (mode == MODE_ZC) or (mode == MODE_ZH):
            return self.hps.seq_width + self.hps.rnn_size
        return self.seq_width  # MODE_Z or MODE_Z_HIDDEN

def get_pi_idx(x, pdf):
    # samples from a categorial distribution
    N = pdf.size
    accumulate = 0
    for i in range(0, N):
        accumulate += pdf[i]
        if x <= accumulate:
            return i
    random_value = np.random.randint(N)
    # print('error with sampling ensemble, returning random', random_value)
    return random_value


def sigmoid(x):
    return 1 / (1 + np.exp(-x))


def rnn_init_state(rnn):
    return rnn.sess.run(rnn.initial_state)


def rnn_next_state(rnn, z, a, prev_state):
    feed = {
        rnn.input_z: z.reshape((1, 1, rnn.hps.seq_width)),
        rnn.input_action: a.reshape((1, 1, rnn.hps.action_width)),
        rnn.input_restart: np.zeros((1, 1)),
        rnn.initial_state: prev_state,
    }
    return rnn.sess.run(rnn.final_state, feed)

def rnn_output(state, z, mode):
    if mode == MODE_ZCH:
        return np.concatenate([z, np.concatenate((state.c, state.h), axis=1)[0]])
    if mode == MODE_ZC:
        return np.concatenate([z, state.c[0]])
    if mode == MODE_ZH:
        return np.concatenate([z, state.h[0]])
    return z  # MODE_Z or MODE_Z_HIDDEN



from __future__ import print_function
import numpy as np
import os

from navrep.models.rnn import reset_graph, sample_hps_params, MDNRNN, get_pi_idx
from navrep.models.vae2d import ConvVAE

# parameters
TEMPERATURE = 0.5
_Z = 32

sequence_z_path = os.path.expanduser(
    "~/navrep/datasets/M/im/corridor_koze_kids_bag_mus_logvars_actions_rewards_dones.npz"
)
rnn_model_path = os.path.expanduser("~/navrep/models/M/imrnn.json")
vae_model_path = os.path.expanduser("~/navrep/models/V/imvae.json")

reset_graph()
imrnn = MDNRNN(sample_hps_params, gpu_mode=False)
imvae = ConvVAE(batch_size=1, is_training=False, channels=3)

imvae.load_json(vae_model_path)
imrnn.load_json(rnn_model_path)

# load sequence image encoding
arrays = np.load(sequence_z_path)
sequence_action = arrays["actions"]
sequence_mu = arrays["mus"]
sequence_logvar = arrays["logvars"]
sequence_z = sequence_mu + np.exp(sequence_logvar / 2.0) * np.random.randn(
    *(sequence_mu.shape)
)
SEQUENCE_LENGTH = len(sequence_mu)

prev_z = sequence_z[0]
prev_z_predicted = sequence_z[0]
prev_action = sequence_action[0]
prev_restart = np.array([0])
rnn_state = imrnn.sess.run(imrnn.zero_state)
for i in range(SEQUENCE_LENGTH):
    print(prev_action)

    feed = {
        imrnn.input_z: np.reshape(prev_z, (1, 1, _Z)),
        imrnn.input_action: np.reshape(prev_action, (1, 1, 3)),
        imrnn.input_restart: np.reshape(prev_restart, (1, 1)),
        imrnn.initial_state: rnn_state,
    }

    [logmix, mean, logstd, logrestart, next_state] = imrnn.sess.run(
        [
            imrnn.out_logmix,
            imrnn.out_mean,
            imrnn.out_logstd,
            imrnn.out_restart_logits,
            imrnn.final_state,
        ],
        feed,
    )

    OUTWIDTH = _Z

    # adjust temperatures
    logmix2 = np.copy(logmix) / TEMPERATURE
    logmix2 -= logmix2.max()
    logmix2 = np.exp(logmix2)
    logmix2 /= logmix2.sum(axis=1).reshape(OUTWIDTH, 1)

    mixture_idx = np.zeros(OUTWIDTH)
    chosen_mean = np.zeros(OUTWIDTH)
    chosen_logstd = np.zeros(OUTWIDTH)
    for j in range(OUTWIDTH):
        idx = get_pi_idx(np.random.rand(), logmix2[j])
        mixture_idx[j] = idx
        chosen_mean[j] = mean[j][idx]
        chosen_logstd[j] = logstd[j][idx]

    rand_gaussian = np.random.randn(OUTWIDTH) * np.sqrt(TEMPERATURE)
    next_z_predicted = chosen_mean + np.exp(chosen_logstd) * rand_gaussian
    if sample_hps_params.differential_z:
        next_z_predicted = prev_z + next_z_predicted  # if in the residual regime

    # take z from sequence (not dream)
    next_z = sequence_z[i]
    prev_action = sequence_action[i]

    next_restart = 0
    done = False
    if logrestart[0] > 0:
        next_restart = 1
        done = True

    im_pred = (imvae.decode(prev_z.reshape(1, _Z))[0] * 255.).astype(np.uint8)
    im_z_pred_pred = (imvae.decode(prev_z_predicted.reshape(1, _Z))[0] * 255.).astype(np.uint8)
    if True:
        import matplotlib.pyplot as plt

        plt.ion()
        plt.figure("images")
        plt.clf()
        fig, (ax1, ax2) = plt.subplots(
            2, 1, num="images"
        )
        ax1.imshow(im_pred)
        ax2.imshow(im_z_pred_pred)
        # update
        plt.pause(0.01)

    prev_z = next_z
    prev_z_predicted = next_z_predicted
    prev_restart = next_restart
    rnn_state = next_state



import os

from navrep.tools.envplayer import EnvPlayer
from navrep.envs.ianenv import IANEnv

if __name__ == "__main__":
    os.environ["CUDA_VISIBLE_DEVICES"] = "-1"  # disable GPU

    env = IANEnv()
    player = EnvPlayer(env, step_by_step=True)



from __future__ import print_function
from datetime import datetime
import numpy as np
import os
import random
import time
import pickle
from pyniel.python_tools.path_tools import make_dir_if_not_exists
import pandas as pd

from navrep.models.tcn import reset_graph, default_hps, MDNTCN
from navrep.tools.commonargs import parse_common_args

common_args, _ = parse_common_args()
VARIANT = common_args.environment
print(common_args)


hps = default_hps()
# hps.batch_size = 100
# hps.max_seq_len = 1000
N_EPOCHS = 10000
START_TIME = datetime.now().strftime("%Y_%m_%d__%H_%M_%S")
_Z = 32

# hps = hps._replace(learning_rate=0.001)

if VARIANT == "ian":
    dataset_folder = os.path.expanduser("~/navrep/datasets/M/ian")
    log_path = os.path.expanduser("~/navrep/logs/M/tcn_train_log_{}.csv".format(START_TIME))
    log_hyperparams_path = os.path.expanduser(
        "~/navrep/logs/M/tcn_train_log_{}.hyperparams.pckl".format(START_TIME))
    model_hyperparams_path = os.path.expanduser("~/navrep/models/M/tcn.hyperparams.pckl")
    model_path = os.path.expanduser("~/navrep/models/M/tcn.json")
    vae_model_path = os.path.expanduser("~/navrep/models/V/vae.json")
if VARIANT == "toy":
    dataset_folder = os.path.expanduser("~/navrep/datasets/M/toy")
    log_path = os.path.expanduser("~/navrep/logs/M/toytcn_train_log_{}.csv".format(START_TIME))
    log_hyperparams_path = os.path.expanduser(
        "~/navrep/logs/M/toytcn_train_log_{}.hyperparams.pckl".format(START_TIME))
    model_hyperparams_path = os.path.expanduser("~/navrep/models/M/toytcn.hyperparams.pckl")
    model_path = os.path.expanduser("~/navrep/models/M/toytcn.json")
    vae_model_path = os.path.expanduser("~/navrep/models/V/toyvae.json")

make_dir_if_not_exists(os.path.dirname(model_path))
make_dir_if_not_exists(os.path.dirname(log_path))

# load preprocessed data
files = []
for dirpath, dirnames, filenames in os.walk(dataset_folder):
    for filename in [f for f in filenames if f.endswith(".npz")]:
        files.append(os.path.join(dirpath, filename))
all_data = []
for path in files:
    arrays = np.load(path)
    all_data.append(
        [
            arrays["mus"],
            arrays["logvars"],
            arrays["actions"],
            arrays["dones"],
            arrays["rewards"],
        ]
    )
n_total_frames = np.sum([mu.shape[0] for mu, _, _, _, _ in all_data])
print("total frames: ", n_total_frames)


reset_graph()
model = MDNTCN(hps)

viewer = None
values_logs = None

start = time.time()
for epoch in range(1, N_EPOCHS + 1):
    #     print('preparing data for epoch', epoch)
    epoch_start = time.time()
    # flatten all sequences into one
    mu_sequence = np.zeros((n_total_frames, _Z), dtype=np.float32)
    logvar_sequence = np.zeros((n_total_frames, _Z), dtype=np.float32)
    action_sequence = np.zeros((n_total_frames, 3), dtype=np.float32)
    done_sequence = np.zeros((n_total_frames, 1), dtype=np.float32)
    reward_sequence = np.zeros((n_total_frames, 1), dtype=np.float32)
    i = 0
    random.shuffle(all_data)
    for mu, logvar, action, done, reward in all_data:
        L = len(mu)
        mu_sequence[i : i + L, :] = mu.reshape(L, _Z)
        logvar_sequence[i : i + L, :] = logvar.reshape(L, _Z)
        action_sequence[i : i + L, :] = action.reshape(L, 3)
        done_sequence[i : i + L, :] = done.reshape(L, 1)
        reward_sequence[i : i + L, :] = reward.reshape(L, 1)
        i += L
    # sample z from  mu and logvar
    z_sequence = mu_sequence + np.exp(logvar_sequence / 2.0) * np.random.randn(
        *(mu_sequence.shape)
    )
    # resize array to be reshapable into sequences and batches
    chunksize = hps.batch_size * hps.max_seq_len  # frames per batch (100'000)
    n_chunks = n_total_frames // chunksize
    # reshape into sequences
    z_sequences = np.reshape(
        z_sequence[: n_chunks * chunksize, :], (-1, hps.max_seq_len, _Z)
    )
    action_sequences = np.reshape(
        action_sequence[: n_chunks * chunksize], (-1, hps.max_seq_len, 3)
    )
    done_sequences = np.reshape(
        done_sequence[: n_chunks * chunksize], (-1, hps.max_seq_len)
    )
    reward_sequences = np.reshape(
        reward_sequence[: n_chunks * chunksize], (-1, hps.max_seq_len)
    )
    num_sequences = len(z_sequences)
    # shuffle
    random_idxs = list(range(num_sequences))
    random.shuffle(random_idxs)
    random_idxs = np.reshape(random_idxs, (-1, hps.batch_size))
    # reshape into batches
    z_batches = z_sequences[random_idxs]
    action_batches = action_sequences[random_idxs]
    done_batches = done_sequences[random_idxs]
    reward_batches = reward_sequences[random_idxs]
    num_batches = len(z_batches)
    # result is of size (n_batches, batch_size, seq_len, ...)
    #     print('number of batches', num_batches)
    batch_create_time = time.time() - epoch_start
    #     print('time taken to create batches', batch_create_time)

    for batch_z, batch_action, batch_done, batch_reward in zip(
        z_batches, action_batches, done_batches, reward_batches
    ):

        if False:  # Visually check that the batch is sound
            from navrep.models.vae2d import ConvVAE
            import matplotlib.pyplot as plt
            from navrep.tools.rings import generate_rings

            reset_graph()
            vae = ConvVAE(batch_size=1, is_training=False)
            vae.load_json(vae_model_path)
            rings_def = generate_rings(64, 64)
            rings_pred = vae.decode(batch_z[0]) * rings_def["rings_to_bool"]
            plt.ion()
            for i, ring in enumerate(rings_pred):
                rings_def["visualize_rings"](ring, scan=None)
                plt.ylim([0, 10])
                plt.title(str(batch_action[0, i]))
                plt.pause(0.1)
            exit()
        if False:
            from navrep.models.vae2d import ConvVAE
            from navrep.tools.render import render_lidar_batch
            from navrep.tools.rings import generate_rings

            reset_graph()
            vae = ConvVAE(batch_size=100, is_training=False)
            vae.load_json(vae_model_path)
            rings_def = generate_rings(64, 64)
            batch_decodings = []
            for i in range(batch_z.shape[1]):  # for each sequence step
                rings_pred = vae.decode(batch_z[:, i]) * rings_def["rings_to_bool"]
                predicted_ranges = rings_def["rings_to_lidar"](rings_pred, 1080)
                batch_decodings.append(predicted_ranges)
            for i, predicted_ranges in enumerate(batch_decodings):
                viewer = render_lidar_batch(
                    predicted_ranges, 0, 2 * np.pi, viewer=viewer
                )
                import pyglet

                filename = "/tmp/frame{:03}.png".format(i)
                pyglet.image.get_buffer_manager().get_color_buffer().save(filename)
                print("image file writen : ", filename)

        step = model.sess.run(model.global_step)
        curr_learning_rate = (hps.learning_rate - hps.min_learning_rate) * (
            hps.decay_rate
        ) ** step + hps.min_learning_rate

        feed = {
            model.batch_z: batch_z,
            model.batch_action: batch_action,
            model.batch_restart: batch_done,
            model.lr: curr_learning_rate,
        }

        (train_cost, rmse, z_cost, r_cost, train_step, _) = model.sess.run(
            [
                model.cost,
                model.rmse,
                model.z_cost,
                model.r_cost,
                model.global_step,
                model.train_op,
            ],
            feed,
        )
        if step % 20 == 0 and step > 0:
            end = time.time()
            time_taken = end - start
            start = time.time()
            output_log = (
                "step: %d, lr: %.6f, cost: %.4f, z_cost: %.4f, r_cost: %.4f, train_time_taken: %.4f"
                % (step, curr_learning_rate, train_cost, z_cost, r_cost, time_taken)
            )
            print(output_log)
            model.save_json(model_path)
            with open(model_hyperparams_path, "wb") as f:
                pickle.dump(hps, f)
            values_log = pd.DataFrame(
                [
                    [
                        step,
                        curr_learning_rate,
                        train_cost,
                        rmse,
                        z_cost,
                        r_cost,
                        time_taken,
                    ]
                ],
                columns=[
                    "step",
                    "lr",
                    "cost",
                    "rmse",
                    "z_cost",
                    "r_cost",
                    "train_time_taken",
                ],
            )
            if values_logs is None:
                values_logs = values_log.copy()
            else:
                values_logs = values_logs.append(values_log, ignore_index=True)
            values_logs.to_csv(log_path)
            with open(log_hyperparams_path, "wb") as f:
                pickle.dump(hps, f)

model.save_json(model_path)



import numpy as np
import torch
import torch.nn as nn
from torch.nn import functional as F

from navrep.models.torchvae import VAE

_A = 3
_G = 2

STATE_NORM_FACTOR = 25.  # maximum typical goal distance, meters

class VAEConfig:
    def __init__(self, z_size):
        self.z_size = z_size

class LSTMConfig:
    n_layer = 2

    def __init__(self, h_size):
        self.h_size = h_size

class VAELSTMConfig:
    pdrop = 0.1

    n_action = _A  # forward v, side v, rot v
    n_states = _G  # goal_x, goal_y (in robot frame)

    def __init__(self, z_size, h_size):
        self.vae = VAEConfig(z_size)
        self.lstm = LSTMConfig(h_size)

class VAELSTM(nn.Module):
    def __init__(self, config, gpu=True):
        super().__init__()

        # input embedding stem
        self.convVAE = VAE(z_dim=config.vae.z_size, gpu=gpu)
        self.action_emb = nn.Linear(config.n_action, config.vae.z_size)
        self.state_emb = nn.Linear(config.n_states, config.vae.z_size)
        self.drop = nn.Dropout(config.pdrop)
        # transformer
        self.lstm = nn.LSTM(config.vae.z_size * 3,  # *3 due to concat (z, action, state)
                            config.lstm.h_size, config.lstm.n_layer,
                            batch_first=True)
        self.lstm_head = nn.Linear(config.lstm.h_size, config.vae.z_size * 2) # *2 (no action pred)
        # decoder head
        self.ln_f = nn.LayerNorm(config.vae.z_size * 2)
        self.z_head = nn.Linear(config.vae.z_size, config.vae.z_size)
        self.state_head = nn.Linear(config.vae.z_size, config.n_states)

        self.config = config
        self.gpu = gpu

        self.apply(self._init_weights)

        print(
            "number of parameters: %e", sum(p.numel() for p in self.parameters())
        )

    def _init_weights(self, module):
        if isinstance(module, (nn.Linear, nn.Embedding)):
            module.weight.data.normal_(mean=0.0, std=0.02)
            if isinstance(module, nn.Linear) and module.bias is not None:
                module.bias.data.zero_()
        elif isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)

    def forward(self, img, state, action, dones, rnn_state, targets=None):
        """
        img: (batch, sequence, CH, W, H) [0, 1]
        action: (batch, sequence, A) [-inf, inf]
        state: (batch, sequence, S) [-inf, inf]
        dones:  (batch, sequence,) {0, 1}
        targets: None or (img_targets, state_targets)
            img_targets: same shape as img
            state_targets: same shape as state
        rnn_state: [h0, c0]

        OUTPUTS
        img_pred: same shape as img
        state_pred: same shape as state
        loss: torch loss
        rnn_state: contents get updated with hn, cn
        """
        b, t, CH, W, H = img.size()
        _, _, A = action.size()
        _, _, S = state.size()

        h0 = rnn_state[0]
        c0 = rnn_state[1]

        # encode embedding with vae
        z, mu, logvar = self.convVAE.encode(
            img.view(b * t, CH, W, H)
        )  # each image maps to a vector
        z_embeddings = z.view(b, t, self.config.vae.z_size)
        state_embeddings = self.state_emb(state.view(b * t, S)).view(b, t, self.config.vae.z_size)
        action_embeddings = self.action_emb(action.view(b * t, A)).view(
            b, t, self.config.vae.z_size
        )
        # forward the prediction model
        x = self.drop(torch.cat([z_embeddings, action_embeddings, state_embeddings], dim=-1))
        x, (hn, cn) = self.lstm(x, (h0, c0))
        x = x.reshape(b * t, self.config.lstm.h_size) # view() doesn't work...
        x = self.lstm_head(x)
        x = self.ln_f(x)
        z_embedding_pred = x[:, :self.config.vae.z_size]
        state_embedding_pred = x[:, self.config.vae.z_size:]
        # store final rnn state
        if rnn_state is not None:
            rnn_state[0] = hn
            rnn_state[1] = cn
        # decode embedding with vae
        state_pred = self.state_head(state_embedding_pred).view(b, t, S)
        z_pred = self.z_head(z_embedding_pred)
        img_rec = self.convVAE.decode(z).view(b, t, CH, W, H)
        img_pred = self.convVAE.decode(z_pred).view(b, t, CH, W, H)

        # if we are given some desired targets also calculate the loss
        loss = None
        if targets is not None:
            img_targets, state_targets = targets
            rec_loss = F.binary_cross_entropy(img_rec, img)  # input-reconstruction loss
            pred_loss = F.binary_cross_entropy(img_pred, img_targets)  # reconstructed prediction loss
            state_loss = F.mse_loss(state_pred, state_targets) / STATE_NORM_FACTOR**2
            KLD = -0.5 * torch.mean(1 + logvar - mu.pow(2) - logvar.exp())  # kullback leibler
            kld_tolerance = 0.5
            kld_weight = 0.001
            KLD = torch.max(KLD, kld_tolerance * torch.ones_like(KLD))
            loss = rec_loss + kld_weight * KLD + pred_loss + state_loss

        return img_pred, state_pred, loss

    def _to_correct_device(self, tensor):
        if self.gpu:
            if torch.cuda.is_available():
                device = torch.cuda.current_device()
                return tensor.to(device)
            else:
                print("WARNING: model created with gpu enabled, but no gpu found")
        return tensor

    def encode(self, img):
        """
        img: numpy (batch, W, H, CH)

        OUTPUTS
        z: (batch, Z)
        """
        b, W, H, CH = img.shape

        img_t = torch.tensor(np.moveaxis(img, -1, 1), dtype=torch.float)
        img_t = self._to_correct_device(img_t)

        z, mu, logvar = self.convVAE.encode(img_t)
        return z.detach().cpu().numpy()

    def encode_mu_logvar(self, img):
        """
        img: numpy (batch, W, H, CH)


        OUTPUTS
        mu: (batch, Z)
        logvar: (batch, Z)
        """
        b, W, H, CH = img.shape

        img_t = torch.tensor(np.moveaxis(img, -1, 1), dtype=torch.float)
        img_t = self._to_correct_device(img_t)

        z, mu, logvar = self.convVAE.encode(img_t)
        mu = mu.detach().cpu().numpy()
        logvar = logvar.detach().cpu().numpy()
        return mu, logvar

    def decode(self, z):
        """
        z: numpy (batch, Z)

        OUTPUTS
        img_rec: (batch, W, H, CH)
        """
        b, Z = z.shape

        z_t = torch.tensor(z, dtype=torch.float)
        z_t = self._to_correct_device(z_t)

        img_rec_t = self.convVAE.decode(z_t) # b, CH, W, H
        img_rec = np.moveaxis(img_rec_t.detach().cpu().numpy(), 1, -1)
        return img_rec

    def get_h(self, gpt_sequence):
        """ for compat with encodedenv
        gpt sequence is a list of dicts, one for each step in the sequence.
        each dict has
        "obs": numpy img (W, H, CH) [0, 1]
        "state": numpy (2,) [-inf, inf]
        "action": numpy (3,) [-inf, inf]

        h: numpy (h_size,)  the hidden state at the end of the sequence
        """
        _b = 1  # batch size
        img = np.array([d["obs"] for d in gpt_sequence])  # t, W, H, CH
        img = np.moveaxis(img, -1, 1)
        img = img.reshape((_b, *img.shape))
        img_t = torch.tensor(img, dtype=torch.float)
        img_t = self._to_correct_device(img_t)
        state = np.array([d["state"] for d in gpt_sequence])  # t, 2
        state = state.reshape((_b, *state.shape))
        state_t = torch.tensor(state, dtype=torch.float)
        state_t = self._to_correct_device(state_t)
        action = np.array([d["action"] for d in gpt_sequence])  # t, 3
        action = action.reshape((_b, *action.shape))
        action_t = torch.tensor(action, dtype=torch.float)
        action_t = self._to_correct_device(action_t)
        dones = np.zeros((_b, len(gpt_sequence), 1))
        dones_t = torch.tensor(dones, dtype=torch.float)
        dones_t = self._to_correct_device(dones_t)
        h0 = torch.randn(self.config.lstm.n_layer, _b, self.config.lstm.h_size)
        c0 = torch.randn(self.config.lstm.n_layer, _b, self.config.lstm.h_size)
        h0 = self._to_correct_device(h0)
        c0 = self._to_correct_device(c0)
        rnn_state_container = [h0, c0]
        self.forward(img_t, state_t, action_t, dones_t, rnn_state=rnn_state_container)
        h = rnn_state_container[0]
        h = h.view(self.config.lstm.n_layer, _b, self.config.lstm.h_size)
        h = h.detach().cpu().numpy()
        h = h[-1, 0, :]  # last layer, only batch
        return h



from __future__ import print_function
import os
import time
import gc
import numpy as np
from pepper_2d_simulator import parse_PepperRLEnv_args, PepperRLEnv, check_PepperRLEnv_args
import gym
from gym import spaces
from pandas import DataFrame
from pkg_resources import resource_filename

from navrep.envs.scenario_list import map_downsampling

PUNISH_SPIN = False

# TODO: this is still a work in progress. @Daniel: Make it into a VecEnv
# from stable_baselines import DummyVecEnv

CROWDMOVE_MAPS = ["crowdmove1", "crowdmove2", "crowdmove3", "crowdmove4", "crowdmove5", "crowdmove6"]

class JointEnv(gym.Env):
    """ This class wraps the IARLenv to make a RL ready simplified environment
    """
    metadata = {'render.modes': ['human']}

    def __init__(self, silent=False, n_robots=10, maps=CROWDMOVE_MAPS, circular_scenario=False,
                 max_episode_length=1000, no_inertia=True, collect_trajectories=False, lidar_legs=False):
        # gym env definition
        super(JointEnv, self).__init__()
        self.action_space = spaces.Box(low=-1, high=1, shape=(n_robots, 3), dtype=np.float32)
        self.observation_space = spaces.Tuple((spaces.Box(low=-np.inf, high=np.inf,
                                                          shape=(n_robots, 4, 1080, 1), dtype=np.float32),
                                              spaces.Box(low=-np.inf, high=np.inf,
                                                         shape=(n_robots, 5), dtype=np.float32)))
        # parameters
        self.silent = silent
        self.n_robots = n_robots
        self.maps = maps
        self.no_inertia = no_inertia
        self.circular_scenario = circular_scenario
        self.lidar_legs = lidar_legs
        # constants
        self.COLLECT_TRAJECTORIES = collect_trajectories
        self.TEMPERATURE = 0.5
        self.DT = 0.2  # should be the same as data rnn was trained with
        # other tools
        self.viewer = None
        self.episode_statistics = DataFrame(
            columns=[
                "total_steps",
                "mode",
                "steps",
                "goal_reached",
                "damage",
                "static_damage",
                "goal",
                "trajectory",
                "other_trajectories",
                "reward",
                "wall_time",
            ])
        self.current_episode_trajectory = []
        self.current_episode_o_trajectories = []
        self.current_episode_goals = None
        self.total_steps = 0
        # environment state variables
        self.reset()

    def step(self, action):
        if action.shape == (3,):
            action = np.repeat(action[None, :], self.n_robots, axis=0)
        self.steps_since_reset += 1
        self.total_steps += 1
        obs, reward, done, infos = self.rlenv.step(action)
        if PUNISH_SPIN:
            reward -= action[2]**2 / 1000.
        self.current_episode_trajectory.append(self.rlenv.virtual_peppers[0].pos * 1.)
        self.current_episode_o_trajectories.append(
            [vp.pos * 1. for vp in self.rlenv.virtual_peppers[1:]])
        if np.any(done):
            traj = None
            o_traj = None
            if self.COLLECT_TRAJECTORIES:
                traj = np.array(self.current_episode_trajectory) * 1.
                o_traj = np.array(self.current_episode_o_trajectories) * 1.
            self.episode_statistics.loc[len(self.episode_statistics)] = [
                self.total_steps,
                self.rlenv.MODE,
                self.steps_since_reset,
                [info["goal_reached"] for info in infos],
                self.rlenv.episode_damage * 1.,
                self.rlenv.episode_damage_from_static * 1.,
                self.current_episode_goals * 1.,
                traj,
                o_traj,
                reward,
                time.time(),
            ]
        return obs, reward, np.any(done), infos

    def reset(self, set_scenario=None):
        try:
            self.rlenv.viewer.close()
        except AttributeError:
            pass
        self._make_env(silent=self.silent, set_scenario=set_scenario)
        obs = self.rlenv.reset()
        self.steps_since_reset = 0
        self.current_episode_trajectory = []
        self.current_episode_o_trajectories = []
        self.current_episode_goals = self.rlenv.agent_goals * 1.
        return obs

    def render(self, *args, **kwargs):
        self.rlenv.render(*args, **kwargs)

    def close(self):
        try:
            self.rlenv.close()
        except AttributeError:
            pass

    def _make_env(self, silent=False, set_scenario=None):
        args, _ = parse_PepperRLEnv_args(args=[])
        args.unmerged_scans = False
        args.continuous = True
        args.no_ros = True
        args.no_legs = not self.lidar_legs
        args.n_agents = self.n_robots
        args.map_folder = resource_filename('asl_pepper_2d_sim_maps', 'maps')
        args.map_name = set_scenario
        if args.map_name is None:
            args.map_name = np.random.choice(self.maps)
        args.map_downsampling_passes = map_downsampling[args.map_name]
        # only do circular scenario in empty map
        args.circular_scenario = False
        if self.circular_scenario and args.map_name == "crowdmove1":
            args.circular_scenario = True
            args.map_downsampling_passes = 0
        check_PepperRLEnv_args(args)
        self.rlenv = PepperRLEnv(args, silent=silent)
        self.rlenv.virtual_peppers[0].NO_INERTIA = self.no_inertia
        gc.collect()

    def _get_viewer(self):
        viewer = None
        try:
            viewer = self.rlenv._get_viewer()
        except AttributeError:
            pass
        return viewer

    def _get_dt(self):
        return self.rlenv.args.dt


if __name__ == "__main__":
    from navrep.tools.envplayer import EnvPlayer
    env = JointEnv()
    player = EnvPlayer(env)



import os
import numpy as np
import pandas as pd
import time
from stable_baselines import PPO2
from matplotlib import pyplot as plt
from pyniel.python_tools.path_tools import make_dir_if_not_exists

from navrep.tools.commonargs import parse_multiproc_args
from navrep.envs.markenv import FIRST_TEST_MAPS
from navrep.envs.markencodedenv import MarkOneEncodedEnv


class MarkOneVMCPolicy(object):
    """
    Compatible with raw lidar envs (IANEnv, MarkEnv),
    takes a (1080 lidar, 5 state) obs as input,
    converts it for gym C,
    outputs (3,) action"""
    def __init__(self):
        self.model_path = os.path.expanduser(
            "~/navrep/models/gym/markoneencodedenv_latest_PPO_ckpt")
        self.encoder_env = MarkOneEncodedEnv(silent=True)
        self.prev_action = np.array([0.,0.,0.])
        self.model = PPO2.load(self.model_path)
        print("Model '{}' loaded".format(self.model_path))

    def act(self, obs):
        vm_obs = self.encoder_env.encoder._encode_obs(obs, self.prev_action)
        action, _states = self.model.predict(vm_obs, deterministic=True)
        action = np.array([action[0], action[1], 0.])  # no rotation
        self.prev_action = action * 1.
        return action

class MarkOneCPolicy(object):
    """
    Compatible with encoded envs (EncodedEnv, MarkOneEncodedEnv, etc...)
    takes a (546) obs as input,
    converts it for gym C,
    outputs (2,) action"""
    def __init__(self):
        self.model_path = os.path.expanduser(
            "~/navrep/models/gym/markoneencodedenv_latest_PPO_ckpt")
        self.model = PPO2.load(self.model_path)
        print("Model '{}' loaded".format(self.model_path))

    def act(self, obs):
        action, _states = self.model.predict(obs, deterministic=True)
        return action

class MarkTwoCPolicy(object):
    """Compatible with encoded envs (EncodedEnv, MarkOneEncodedEnv, etc...)
    takes a (546) obs as input,
    converts it for gym C,
    outputs (2,) action"""
    def __init__(self):
        self.model_path = os.path.expanduser(
            "~/navrep/models/gym/marktwoencodedenv_latest_PPO_ckpt")
        self.model = PPO2.load(self.model_path)
        print("Model '{}' loaded".format(self.model_path))

    def act(self, obs):
        action, _states = self.model.predict(obs, deterministic=True)
        return action

def markeval(policy, env, n_episodes=None,
             subset_index=0, n_subsets=1,
             render=False,
             ):
    if n_episodes is None:
        n_episodes = 100

    episode_ids = np.arange(n_episodes)
    if n_subsets > 1:  # when multiprocessing
        episode_ids = np.array_split(episode_ids, n_subsets)[subset_index]

    markeval_dir = "~/navrep/eval/markeval"
    stats_dir = os.path.join(os.path.expanduser(markeval_dir), os.path.basename(policy.model_path))
    make_dir_if_not_exists(stats_dir)
    stats_path = os.path.join(stats_dir, "episode_statistics_{}_to_{}.pckl".format(
        episode_ids[0], episode_ids[-1]))

    # Test for n episodes
    obs = env.reset()
    # run policy
    for i in episode_ids:
        while True:
            action = policy.act(obs)
            obs, rewards, done, info = env.step(action)
            if render:
                env.render()
            if done:
                env.reset()
                break
        print("Simulating episode {}/{}".format(i, n_episodes), end="\r")

    S = env.episode_statistics
    S.to_pickle(stats_path)

    if subset_index == 0:
        while True:
            all_S = load_markeval_statistics(stats_dir)
            if len(all_S) < n_episodes:
                time.sleep(1.)
            else:
                break

        plot_markeval_statistics(all_S, env)

    return env.episode_statistics

def clear_markeval_statistics(stats_dir):
    files = []
    for dirpath, dirnames, filenames in os.walk(stats_dir):
        for filename in [f for f in filenames if f.endswith(".pckl")]:
            files.append(os.path.join(dirpath, filename))
    files = sorted(files)
    if files:
        key = input("Directory {} already contains {} eval statistic files. Delete? [y/N]".format(
            stats_dir, len(files)))
        if key.lower() in ['y', 'yes']:
            for file in files:
                os.remove(file)
                print("{} removed.".format(file))
        else:
            raise ValueError("No way forward. (files already exist)")

def load_markeval_statistics(stats_dir):
    files = []
    for dirpath, dirnames, filenames in os.walk(stats_dir):
        for filename in [f for f in filenames if f.endswith(".pckl")]:
            files.append(os.path.join(dirpath, filename))
    files = sorted(files)
    S_parts = [pd.read_pickle(file) for file in files]
    all_S = pd.concat(S_parts, ignore_index=True)
    return all_S

def plot_markeval_statistics(S, env):
    plt.ion()
    scenarios = sorted(list(set(S["scenario"].values)))
    fig, axes = plt.subplots(len(scenarios), 1)
    axes = np.array(axes).reshape((-1))
    for i, scenario in enumerate(scenarios):
        is_scenario = S["scenario"].values == scenario
        scenario_S = S[is_scenario]
        # print results
        goal_reached = scenario_S["goal_reached"].values
        goal_reached_perc = np.sum(goal_reached) * 100. / len(goal_reached)
        timed_out = scenario_S["timed_out"].values
        timed_out_perc = np.sum(timed_out) * 100. / len(timed_out)
        collisioned_out = scenario_S["collisioned_out"].values
        collisioned_out_perc = np.sum(collisioned_out) * 100. / len(collisioned_out)
        print("{} -----------------------------------".format(scenario))
        print("Goal reached: {}%".format(goal_reached_perc))
        print("Timed out: {}%".format(timed_out_perc))
        print("Collisioned out: {}%".format(collisioned_out_perc))

        # plot result
        trajectories = scenario_S["trajectory"]
        goals = scenario_S["goal"]

        ax = axes[i]

        # TODO make this a map2d utility function
        # TODO get scenario map
        env.reset(set_scenario=scenario)
        contours = env.rlenv.virtual_peppers[0].map2d.as_closed_obst_vertices()
        for c in contours:
            cplus = np.concatenate((c, c[:1, :]), axis=0)
            ax.plot(cplus[:,0], cplus[:,1], color='k')
        # ^^^^

        for t, g, s in zip(trajectories, goals, goal_reached):
            line_color = "blue" if s else "orange"
            ax.plot(t[:,0], t[:,1], color=line_color)
            ax.scatter(t[0,0], t[0,1], color="green")
            ax.scatter(g[0], g[1], color="red")
        ax.axis("equal")

    goal_reached = S["goal_reached"].values
    goal_reached_perc = np.sum(goal_reached) * 100. / len(goal_reached)
    timed_out = S["timed_out"].values
    timed_out_perc = np.sum(timed_out) * 100. / len(timed_out)
    collisioned_out = S["collisioned_out"].values
    collisioned_out_perc = np.sum(collisioned_out) * 100. / len(collisioned_out)
    print("All scenarios -----------------------------------")
    print("Goal reached: {}%".format(goal_reached_perc))
    print("Timed out: {}%".format(timed_out_perc))
    print("Collisioned out: {}%".format(collisioned_out_perc))
    plt.pause(1.)


if __name__ == "__main__":
    args, _ = parse_multiproc_args()

    # example usage
    env = MarkOneEncodedEnv(silent=True, maps=FIRST_TEST_MAPS)
    policy = MarkOneCPolicy()
    S = markeval(policy, env,
                 n_episodes=args.n,
                 subset_index=args.subproc_id, n_subsets=args.n_subprocs,
                 render=args.render)



import os

from navrep.tools.envplayer import EnvPlayer
from navrep.envs.markonedreamenv import MarkOneDreamEnv

if __name__ == "__main__":
    os.environ["CUDA_VISIBLE_DEVICES"] = "-1"  # disable GPU

    denv = MarkOneDreamEnv(temperature=0.)
    player = EnvPlayer(denv)



from gym import spaces
import numpy as np
import os

from navrep.envs.navreptrainenv import NavRepTrainEnv
from navrep.envs.encodedenv import EnvEncoder

class NavRepTrainEncoder(EnvEncoder):
    def __init__(self, backend, encoding,
                 gpu=False, encoder_to_share_model_with=None):
        super(NavRepTrainEncoder, self).__init__(
            backend, encoding,
            rnn_model_path=os.path.expanduser("~/navrep/models/M/navreptrainrnn.json"),
            rnn1d_model_path=os.path.expanduser("~/navrep/models/M/navreptrainrnn1d.json"),
            vae_model_path=os.path.expanduser("~/navrep/models/V/navreptrainvae.json"),
            vae1d_model_path=os.path.expanduser("~/navrep/models/V/navreptrainvae1d.json"),
            gpt_model_path=os.path.expanduser("~/navrep/models/W/navreptraingpt"),
            gpt1d_model_path=os.path.expanduser("~/navrep/models/W/navreptraingpt1d"),
            vae1dlstm_model_path=os.path.expanduser("~/navrep/models/W/navreptrainvae1dlstm"),
            vaelstm_model_path=os.path.expanduser("~/navrep/models/W/navreptrainvaelstm"),
            gpu=gpu,
            encoder_to_share_model_with=None,
        )


class NavRepTrainEncodedEnv(NavRepTrainEnv):
    """ takes a (2) action as input
    outputs encoded obs (546) """
    def __init__(self, backend, encoding,
                 scenario='test', silent=False, adaptive=True,
                 gpu=False, shared_encoder=None, encoder=None):
        if encoder is None:
            encoder = NavRepTrainEncoder(backend, encoding,
                                         gpu=gpu, encoder_to_share_model_with=shared_encoder)
        self.encoder = encoder
        super(NavRepTrainEncodedEnv, self).__init__(scenario=scenario, silent=silent, adaptive=adaptive,
                                                    legacy_mode=False)
        self.action_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        self.observation_space = self.encoder.observation_space

    def step(self, action):
        action = np.array([action[0], action[1], 0.])  # no rotation
        obs, reward, done, info = super(NavRepTrainEncodedEnv, self).step(action)
        h = self.encoder._encode_obs(obs, action)
        return h, reward, done, info

    def reset(self, *args, **kwargs):
        self.encoder.reset()
        obs = super(NavRepTrainEncodedEnv, self).reset(*args, **kwargs)
        h = self.encoder._encode_obs(obs, np.array([0,0,0]))
        return h

    def close(self):
        super(NavRepTrainEncodedEnv, self).close()
        self.encoder.close()

    def render(self, mode="human", close=False, save_to_file=False,
               robocentric=False, render_decoded_scan=True):
        decoded_scan = None
        if render_decoded_scan:
            decoded_scan = self.encoder._get_last_decoded_scan()
        super(NavRepTrainEncodedEnv, self).render(
            mode=mode, close=close, lidar_scan_override=decoded_scan, save_to_file=save_to_file,
            robocentric=robocentric)
        if mode == "rings":
            self.encoder._render_rings(close=close, save_to_file=save_to_file)
        if mode == "polar":
            self.encoder._render_rings_polar(close=close, save_to_file=save_to_file)



from __future__ import print_function
import numpy as np
import os
from navrep.models.vae2d import ConvVAE, reset_graph
from navrep.tools.data_extraction import rosbag_to_image_dataset

DEBUG_PLOTTING = True

# Parameters for training
batch_size = 100
NUM_EPOCH = 1000  # 10
DATA_DIR = "record"
HOME = os.path.expanduser("~")

model_save_dir = HOME + "/navrep/models/V"
model_save_path = os.path.join(model_save_dir, "imvae.json")
if not os.path.exists(model_save_dir):
    os.makedirs(model_save_dir)

# create network
reset_graph()
vae = ConvVAE(batch_size=batch_size, is_training=True, reuse=False, channels=3,)

# create training dataset
dataset, _, _, _ = rosbag_to_image_dataset("~/rosbags/openlab_rosbags/corridor_koze_kids.bag")
if len(dataset) == 0:
    raise ValueError("no images found, exiting")

# split into batches:
total_length = len(dataset)
num_batches = int(np.floor(total_length / batch_size))


# train loop:
print("train", "step", "loss", "recon_loss", "kl_loss")
for epoch in range(NUM_EPOCH):
    np.random.shuffle(dataset)
    for idx in range(num_batches):
        batch = dataset[idx * batch_size : (idx + 1) * batch_size]
        images = batch

        obs = images.astype(float) / 255.0

        feed = {
            vae.x: obs,
        }

        (train_loss, r_loss, kl_loss, train_step, _) = vae.sess.run(
            [vae.loss, vae.r_loss, vae.kl_loss, vae.global_step, vae.train_op], feed
        )

        if (train_step + 1) % 500 == 0:
            if DEBUG_PLOTTING:
                from matplotlib import pyplot as plt

                #           plt.ion()
                plt.figure("training_status")
                plt.clf()
                plt.suptitle("training step {}".format(train_step))
                f, (ax1, ax2) = plt.subplots(2, 1, num="training_status")
                ax1.imshow(images[0])
                ax2.imshow((vae.encode_decode(obs)[0] * 255.0).astype(np.uint8))
                plt.savefig("/tmp/imvae_step{:07}.png".format(train_step))
            #           plt.pause(0.01)
            print("step", (train_step + 1), train_loss, r_loss, kl_loss)
            #     if ((train_step+1) % 5000 == 0):
            vae.save_json(model_save_path)

# finished, final model:
vae.save_json(model_save_path)



import numpy
from setuptools import setup
from Cython.Build import cythonize


setup(
    name="navrep",
    description='navigation representations',
    author='Daniel Dugas',
    version='0.0.13',
    # doesn't work with pip install -e
    package_dir={'frame_msgs': "external/frame_msgs", 'tf_bag': "external/tf_bag"},
    packages=["navrep",
              "navrep.tools",
              "navrep.envs",
              "navrep.models",
              "navrep.scripts",
              # external modules bundled in here for convenience
              "frame_msgs",
              "frame_msgs.msg",
              "tf_bag"],
    # packages=["navrep"],
    ext_modules=cythonize("crings/crings.pyx", annotate=True),
    python_requires='>=3.6, <3.7',
    install_requires=[
        'numpy', 'matplotlib', 'Cython', 'rospkg', 'pyyaml', 'numba', 'ipython', 'flake8', 'pylint', 'rich',
        'tqdm',  # progress bar
        'pandas==1.1.0', # for compatibility with saved files
        'stable-baselines==2.10.0', 'cloudpickle==1.6.0', # try 1.3.0 if fails to load older C models
        'torch==1.6.0', 'torchvision==0.3.0',  # used by CADRL, minGPT
        # 'tensorflow-gpu==1.13.2',
        'tensorflow==1.13.2',
        'keras==2.3.1', 'catkin_pkg', # needed by deep_social_planner
        'pyIAN', 'asl-pepper-2d-sim', 'soadrl-crowdnav',
        'pyniel', 'pymap2d', 'pylidar2d', 'pyrvo2-danieldugas', 'pyrangelibc-danieldugas',
    ],
    include_dirs=[numpy.get_include()],
    package_data={'navrep': ['maps/*']},
)



from navrep.envs.navreptrainenv import NavRepTrainEnv
from navrep.tools.commonargs import parse_common_args
from navrep.scripts.test_navrep import run_test_episodes

class LuciaPolicy(object):
    """ legacy SOADRL policy from lucia's paper, takes in agents state, local map """
    def __init__(self, env):
        self.env = env

    def act(self, obs):
        state, local_map = obs
        return self.env.soadrl_sim.robot.act(state, local_map)


if __name__ == '__main__':
    args, _ = parse_common_args()

    env = NavRepTrainEnv(silent=True, scenario='test', legacy_mode=True)
    policy = LuciaPolicy(env)

    run_test_episodes(env, policy, render=args.render)



from navrep.tools.envplayer import EnvPlayer
from navrep.envs.archiveenv import ArchiveEnv
from navrep.tools.commonargs import parse_common_args

if __name__ == "__main__":
    args, _ = parse_common_args()

    if args.environment is None:
        args.environment = "irl"

    dataset_folder = "~/navrep/datasets/V/{}".format(args.environment)

    env = ArchiveEnv(dataset_folder)
    player = EnvPlayer(env, step_by_step=True)



import os
import pandas as pd

from stable_baselines import PPO2
from stable_baselines.common.vec_env import SubprocVecEnv, DummyVecEnv

from navrep.envs.navreptrainencodedenv import NavRepTrainEncodedEnv, NavRepTrainEncoder
from navrep.tools.sb_eval_callback import NavrepEvalCallback
from navrep.tools.commonargs import parse_common_args

if __name__ == "__main__":
    args, _ = parse_common_args()

    shared_encoder = NavRepTrainEncoder(args.backend, args.encoding, gpu=not args.no_gpu)
    _Z = shared_encoder._Z
    _H = shared_encoder._H

    DIR = os.path.expanduser("~/navrep/models/gym")
    LOGDIR = os.path.expanduser("~/navrep/logs/gym")
    if args.dry_run:
        DIR = "/tmp/navrep/models/gym"
        LOGDIR = "/tmp/navrep/logs/gym"
#     START_TIME = datetime.now().strftime("%Y_%m_%d__%H_%M_%S")
    START_TIME = "2020_10_15__12_02_22"
    ENCODER_ARCH = "_{}_{}_V{}M{}".format(args.backend, args.encoding, _Z, _H)
    LOGNAME = "navreptrainencodedenv_" + START_TIME + "_PPO" + ENCODER_ARCH
    LOGPATH = os.path.join(LOGDIR, LOGNAME + ".csv")
    MODELPATH = os.path.join(DIR, LOGNAME + "_ckpt")
    MODELPATH2 = os.path.join(DIR, "navreptrainencodedenv_latest_PPO_ckpt")
    if not os.path.exists(DIR):
        os.makedirs(DIR)
    if not os.path.exists(LOGDIR):
        os.makedirs(LOGDIR)

    MILLION = 1000000
    TRAIN_STEPS = args.n
    if TRAIN_STEPS is None:
        TRAIN_STEPS = 60 * MILLION

    N_ENVS = 6
    if args.debug:
        if args.backend in ["VAE_LSTM", "VAE1D_LSTM"]:
            # can't share encoder for VAE_LSTM as it contains RNN state internally
            shared_encoder = None
        env = DummyVecEnv([lambda: NavRepTrainEncodedEnv(args.backend, args.encoding,
                                                         silent=True, scenario='train',
                                                         gpu=not args.no_gpu,
                                                         shared_encoder=shared_encoder)]*N_ENVS)
    else:
        env = SubprocVecEnv([lambda: NavRepTrainEncodedEnv(args.backend, args.encoding,
                                                           silent=True, scenario='train',
                                                           gpu=not args.no_gpu)]*N_ENVS,
                            start_method='spawn')
    eval_env = NavRepTrainEncodedEnv(args.backend, args.encoding,
                                     silent=True, scenario='train', gpu=not args.no_gpu)
    def test_env_fn():  # noqa
        return NavRepTrainEncodedEnv(args.backend, args.encoding,
                                     silent=True, scenario='test', gpu=not args.no_gpu)
    cb = NavrepEvalCallback(eval_env, test_env_fn=test_env_fn,
                            logpath=LOGPATH, savepath=MODELPATH, verbose=1)
#     model = PPO2(MlpPolicy, env, verbose=0)  #
    S = pd.read_csv(LOGPATH, index_col=0)
    cb.eval_env.episode_statistics = S
    model = PPO2.load(MODELPATH)
    model.set_env(env)
    print(S)
    model.learn(total_timesteps=TRAIN_STEPS+1, callback=cb)
    obs = env.reset()

    model.save(MODELPATH)
    model.save(MODELPATH2)
    print("Model '{}' saved".format(MODELPATH))

    del model
    env.close()

    model = PPO2.load(MODELPATH)

    env = NavRepTrainEncodedEnv(args.backend, args.encoding,
                                silent=True, scenario='train')
    obs = env.reset()
    for i in range(512):
        action, _states = model.predict(obs, deterministic=True)
        obs, _, done, _ = env.step(action)
        if done:
            env.reset()
#         env.render()

    print("exiting.")
    exit()



import os

from navrep.envs.markenv import MarkEnv
from navrep.scripts.test_mark_common import load_markeval_statistics, plot_markeval_statistics


if __name__ == "__main__":
    os.environ["CUDA_VISIBLE_DEVICES"] = "-1"  # disable GPU

    # example usage
    env = MarkEnv(silent=True)  # for plotting only
    stats_dir = os.path.expanduser("~/navrep/eval/markeval/markoneencodedenv_2020_Jul_09__10_47_32_PPO_ckpt")
    S = load_markeval_statistics(stats_dir)
    plot_markeval_statistics(S, env)



from __future__ import print_function
import time
import gc
import numpy as np
from pepper_2d_iarlenv import parse_iaenv_args, IARLEnv, check_iaenv_args
import gym
from gym import spaces
from pandas import DataFrame

from navrep.envs.scenario_list import set_rl_scenario

PUNISH_SPIN = True

class IANEnv(gym.Env):
    """ This class wraps the IARLenv to make a RL ready simplified environment

        Action space:
            (x velocity, y velocity, theta velocity), all in [m/s]
        Observation space:
            (scan, robotstate)
            scan:
                360 degree lidar scan, 1080 rays,
                with min angle of 0 (straight forwards), max angle of 2pi. [m]
            robotstate:
                (gx [m], gy [m], vx [m/s], vy [m/s], vth [rad/s]) - all in robot frame
    """
    metadata = {'render.modes': ['human']}

    def __init__(self, silent=False, max_episode_length=1000, collect_trajectories=False):
        # gym env definition
        super(IANEnv, self).__init__()
        self.action_space = spaces.Box(low=-1, high=1, shape=(3,), dtype=np.float32)
        self.observation_space = spaces.Tuple((
            spaces.Box(low=-np.inf, high=np.inf, shape=(1080,), dtype=np.float32),
            spaces.Box(low=-np.inf, high=np.inf, shape=(5,), dtype=np.float32),
        ))
        # parameters
        self.silent = silent
        self.max_episode_length = max_episode_length
        self.max_episode_damage = 5.  # leads to ~0.5% of the original reward
        # constants
        self.COLLECT_TRAJECTORIES = collect_trajectories
        self.TEMPERATURE = 0.5
        self.DT = 0.2  # should be the same as data rnn was trained with
        # other tools
        self.viewer = None
        self.episode_statistics = DataFrame(
            columns=[
                "total_steps",
                "scenario",
                "damage",
                "steps",
                "goal_reached",
                "timed_out",
                "collisioned_out",
                "static_damage_proportion",
                "goal",
                "trajectory",
                "other_trajectories",
                "reward",
                "wall_time",
            ])
        self.current_episode_trajectory = []
        self.current_episode_o_trajectories = []
        self.current_episode_goal = None
        self.total_steps = 0
        # environment state variables
#         self.reset()

    def step(self, action):
        self.steps_since_reset += 1
        self.total_steps += 1
#         action = np.array([action[0], action[1], 0.])  # no rotation
        obs, reward, done, info = self.iarlenv.step(action, ONLY_FOR_AGENT_0=True)
        timed_out = self.iarlenv.rlenv.episode_step[0] >= self.max_episode_length
        collisioned_out = self.iarlenv.rlenv.episode_damage[0] >= self.max_episode_damage
        if self.iarlenv.rlenv.episode_damage[0] == 0:
            static_damage_proportion = np.nan
        else:
            static_damage_proportion = self.iarlenv.rlenv.episode_damage_from_static[0] / (
                self.iarlenv.rlenv.episode_damage[0])
        if info["goal_reached"]:
            done = True
        if timed_out:
            done = True
        if collisioned_out:
            done = True
        if PUNISH_SPIN:
            reward -= action[2]**2 / 1000.
        self.current_episode_trajectory.append(self.iarlenv.rlenv.virtual_peppers[0].pos * 1.)
        self.current_episode_o_trajectories.append(
            [vp.pos * 1. for vp in self.iarlenv.rlenv.virtual_peppers[1:]])
        if done:
            traj = None
            o_traj = None
            if self.COLLECT_TRAJECTORIES:
                traj = np.array(self.current_episode_trajectory) * 1.
                o_traj = np.array(self.current_episode_o_trajectories) * 1.
            self.episode_statistics.loc[len(self.episode_statistics)] = [
                self.total_steps,
                self.iarlenv.args.scenario,
                self.iarlenv.rlenv.episode_damage[0] * 1.,
                self.steps_since_reset,
                info["goal_reached"],
                timed_out,
                collisioned_out,
                static_damage_proportion,
                self.current_episode_goal * 1.,
                traj,
                o_traj,
                reward,
                time.time(),
            ]
        scan = obs[0]
        obs = (scan[-1][:, 0], obs[1])  # latest scan only (buffer, ray, channel)
        return obs, reward, done, info

    def reset(self, set_scenario=None):
        try:
            self.iarlenv.rlenv.viewer.close()
        except AttributeError:
            pass
        self._make_env(silent=self.silent, set_scenario=set_scenario)
        self.iarlenv.agents_pos0[0,2] = np.random.random() * 2 * np.pi
        obs = self.iarlenv.reset(ONLY_FOR_AGENT_0=True)
        self.steps_since_reset = 0
        self.current_episode_trajectory = []
        self.current_episode_o_trajectories = []
        self.current_episode_goal = self.iarlenv.rlenv.agent_goals[0] * 1.
        obs = (obs[0][-1,:,0], obs[1])
        return obs

    def render(self, *args, **kwargs):
        self.iarlenv.render(*args, **kwargs)

    def close(self):
        try:
            self.iarlenv.close()
        except AttributeError:
            pass

    def _make_env(self, silent=False, set_scenario=None):
        args = parse_iaenv_args(args=[])
        args.unmerged_scans = False
        args.continuous = True
        args.naive_plan = True
        args.no_ros = True
        args.no_pass_through = True
        set_rl_scenario(args, scenario_name=set_scenario)  # pick a random scenario
        check_iaenv_args(args)
        self.iarlenv = IARLEnv(args, silent=silent)
        self.iarlenv.rlenv.virtual_peppers[0].NO_INERTIA = True
        gc.collect()

    def _get_viewer(self):
        viewer = None
        try:
            viewer = self.iarlenv._get_viewer()
        except AttributeError:
            pass
        return viewer

    def _get_dt(self):
        return self.iarlenv.args.dt


if __name__ == "__main__":
    from navrep.tools.envplayer import EnvPlayer

    env = IANEnv()
    player = EnvPlayer(env)



from __future__ import print_function
import numpy as np
import tensorflow as tf
from collections import namedtuple
import json
import pickle

HyperParams = namedtuple(
    "HyperParams",
    [
        "max_seq_len",
        "seq_width",
        "h_size",
        "batch_size",
        "grad_clip",
        "num_mixture",
        "restart_factor",
        "learning_rate",
        "decay_rate",
        "min_learning_rate",
        "use_layer_norm",
        "dropout_prob",
        "is_training",
        "differential_z",
        "layer_channels",
        "layer_dilations",
        "kernel_size",
    ],
)


def default_hps():
    return HyperParams(
        max_seq_len=1000,  # train on sequences of 1000
        seq_width=32,  # width of our data (32)
        h_size=512,  # number of rnn cells
        batch_size=100,  # minibatch sizes
        grad_clip=1.0,
        num_mixture=5,  # number of mixtures in MDN
        restart_factor=10.0,  # factor of importance for restart=1 rare case for loss.
        learning_rate=0.001,
        decay_rate=0.99999,
        min_learning_rate=0.00001,
        use_layer_norm=0,  # set this to 1 to get more stable results (less chance of NaN), but slower
        dropout_prob=0.05,
        is_training=1,
        differential_z=1,
        layer_channels=[32, 32, 32, 512], # [32, 32, 32, 32, 32, 32, 32, 512],  # last channel size is h_size
        layer_dilations=[1, 2, 4, 8], # [1, 2, 4, 8, 16, 32, 64, 128],
        kernel_size=3, # 7,
    )


default_hps_params = default_hps()
sample_hps_params = default_hps_params._replace(
    batch_size=1, is_training=0
)


def reset_graph():
    if "sess" in globals() and sess:  # noqa
        sess.close()  # noqa
    tf.reset_default_graph()


# MDN-TCN model tailored for doomrnn
class MDNTCN:
    def __init__(self, hps, gpu_mode=True, reuse=False):
        self.hps = hps
        if not gpu_mode:
            with tf.device("/cpu:0"):
                print("model using cpu")
                self.g = tf.Graph()
                with self.g.as_default():
                    self.build_model(hps)
        else:
            print("model using gpu")
            self.g = tf.Graph()
            with self.g.as_default():
                self.build_model(hps)
        self.init_session()

    def build_model(self, hps):

        self.num_mixture = hps.num_mixture
        KMIX = self.num_mixture  # 5 mixtures
        ACTION_WIDTH = 3
        LENGTH = self.hps.max_seq_len - 1  # 999 timesteps

        if hps.is_training:
            self.global_step = tf.Variable(0, name="global_step", trainable=False)

        self.sequence_lengths = LENGTH  # assume every sample has same length.
        self.batch_z = tf.placeholder(
            dtype=tf.float32,
            shape=[self.hps.batch_size, self.hps.max_seq_len, hps.seq_width],
        )
        self.batch_action = tf.placeholder(
            dtype=tf.float32,
            shape=[self.hps.batch_size, self.hps.max_seq_len, ACTION_WIDTH],
        )
        self.batch_restart = tf.placeholder(
            dtype=tf.float32, shape=[self.hps.batch_size, self.hps.max_seq_len]
        )
        # TODO: add latent goal info

        self.input_z = self.batch_z[:, :LENGTH, :]
        self.input_action = self.batch_action[:, :LENGTH, :]
        self.input_restart = self.batch_restart[:, :LENGTH]

        if hps.differential_z:
            self.target_z = (
                self.batch_z[:, 1:, :] - self.batch_z[:, :LENGTH, :]
            )  # delta Z prediction
        else:
            self.target_z = self.batch_z[:, 1:, :]
        self.target_restart = self.batch_restart[:, 1:]

        self.input_seq = tf.concat(
            [
                self.input_z,
                self.input_action,
                tf.reshape(self.input_restart, [self.hps.batch_size, LENGTH, 1]),
            ],
            axis=2,
        )

        # OUTPUT = TCN(INPUT) --------------------------------------------------------
        with tf.variable_scope("MDNTCN"):
            x = self.input_seq  # [batch, seq_length, 32 + 3 + 1]

            padding = "causal"
            dropout_rate = self.hps.dropout_prob
            kernel_size = self.hps.kernel_size
            layer_channels = self.hps.layer_channels
            layer_dilations = self.hps.layer_dilations
            for nb_filters, dilation_rate in zip(layer_channels, layer_dilations):
                init = tf.keras.initializers.RandomNormal(mean=0.0, stddev=0.01)

                # block1
                conv1 = tf.keras.layers.Conv1D(
                    filters=nb_filters,
                    kernel_size=kernel_size,
                    dilation_rate=dilation_rate,
                    padding=padding,
                    kernel_initializer=init,
                )
                batch1 = tf.keras.layers.BatchNormalization(axis=-1)
                ac1 = tf.keras.layers.Activation("relu")
                drop1 = tf.keras.layers.Dropout(rate=dropout_rate)

                # block2
                conv2 = tf.keras.layers.Conv1D(
                    filters=nb_filters,
                    kernel_size=kernel_size,
                    dilation_rate=dilation_rate,
                    padding=padding,
                    kernel_initializer=init,
                )
                batch2 = tf.keras.layers.BatchNormalization(axis=-1)
                ac2 = tf.keras.layers.Activation("relu")
                drop2 = tf.keras.layers.Dropout(rate=dropout_rate)

                downsample = tf.keras.layers.Conv1D(
                    filters=nb_filters,
                    kernel_size=1,
                    padding="same",
                    kernel_initializer=init,
                )
                ac3 = tf.keras.layers.Activation("relu")

                prev_x = x
                x = conv1(x)
                x = batch1(x)
                x = ac1(x)
                x = drop1(x) if self.hps.is_training else x

                x = conv2(x)
                x = batch2(x)
                x = ac2(x)
                x = drop2(x) if self.hps.is_training else x

                if prev_x.shape[-1] != x.shape[-1]:  # match the dimention
                    prev_x = downsample(prev_x)
                assert prev_x.shape == x.shape

                x = ac3(prev_x + x)  # skip connection

            output = x
        # ----------------------------------------------------------------------------
        # output = h = [batch, length, hps.h_size]
        self.h = output

        output = tf.reshape(output, [-1, self.hps.h_size])
        # now [batch * seq_length, hps.h_size]

        NOUT = hps.seq_width * KMIX * 3 + 1  # plus 1 to predict the restart state.

        with tf.variable_scope("MDNTCN"):
            output_w = tf.get_variable("output_w", [self.hps.h_size, NOUT])
            output_b = tf.get_variable("output_b", [NOUT])

        output = tf.reshape(output, [-1, hps.h_size])
        output = tf.nn.xw_plus_b(output, output_w, output_b)
        # now [batch * seq_length, 32 * KMIX * 3 + 1]

        self.out_restart_logits = output[:, 0]  # [batch * seq_length,]
        output = output[:, 1:]

        output = tf.reshape(output, [-1, KMIX * 3])
        # now [batch * seq_length * 32, KMIX * 3]

        logSqrtTwoPI = np.log(np.sqrt(2.0 * np.pi))

        def tf_lognormal(y, mean, logstd):
            return -0.5 * ((y - mean) / tf.exp(logstd)) ** 2 - logstd - logSqrtTwoPI

        def get_lossfunc(logmix, mean, logstd, y):
            v = logmix + tf_lognormal(y, mean, logstd)
            v = tf.reduce_logsumexp(v, 1, keepdims=True)
            return -tf.reduce_mean(v)

        def get_mdn_coef(output):
            logmix, mean, logstd = tf.split(output, 3, 1)
            logmix = logmix - tf.reduce_logsumexp(logmix, 1, keepdims=True)
            return logmix, mean, logstd

        out_logmix, out_mean, out_logstd = get_mdn_coef(output)
        # each [batch * seq_length * 32, 1]

        self.out_logmix = out_logmix
        self.out_mean = out_mean
        self.out_logstd = out_logstd

        # reshape target data so that it is compatible with prediction shape
        flat_target_data = tf.reshape(self.target_z, [-1, 1])

        lossfunc = get_lossfunc(out_logmix, out_mean, out_logstd, flat_target_data)

        self.rmse = tf.sqrt(tf.reduce_mean(tf.square(out_mean - flat_target_data)))
        self.z_cost = tf.reduce_mean(lossfunc)

        flat_target_restart = tf.reshape(self.target_restart, [-1, 1])

        self.r_cost = tf.nn.sigmoid_cross_entropy_with_logits(
            labels=flat_target_restart,
            logits=tf.reshape(self.out_restart_logits, [-1, 1]),
        )

        factor = tf.ones_like(self.r_cost) + flat_target_restart * (
            self.hps.restart_factor - 1.0
        )

        self.r_cost = tf.reduce_mean(tf.multiply(factor, self.r_cost))

        #         self.cost = self.rmse + self.r_cost
        self.cost = self.z_cost + self.r_cost

        if self.hps.is_training == 1:
            self.lr = tf.Variable(self.hps.learning_rate, trainable=False)
            optimizer = tf.train.AdamOptimizer(self.lr)

            gvs = optimizer.compute_gradients(self.cost)
            capped_gvs = [
                (tf.clip_by_value(grad, -self.hps.grad_clip, self.hps.grad_clip), var)
                for grad, var in gvs
            ]
            self.train_op = optimizer.apply_gradients(
                capped_gvs, global_step=self.global_step, name="train_step"
            )

        # initialize vars
        self.init = tf.global_variables_initializer()

        t_vars = tf.trainable_variables()
        self.assign_ops = {}
        for var in t_vars:
            if var.name.startswith("MDNTCN"):
                pshape = var.get_shape()
                pl = tf.placeholder(tf.float32, pshape, var.name[:-2] + "_placeholder")
                assign_op = var.assign(pl)
                self.assign_ops[var] = (assign_op, pl)

    def init_session(self):
        """Launch TensorFlow session and initialize variables"""
        self.sess = tf.Session(graph=self.g)
        self.sess.run(self.init)

    def close_sess(self):
        """ Close TensorFlow session """
        self.sess.close()

    def get_model_params(self):
        # get trainable params.
        model_names = []
        model_params = []
        model_shapes = []
        with self.g.as_default():
            t_vars = tf.trainable_variables()
            for var in t_vars:
                if var.name.startswith("MDNTCN"):
                    param_name = var.name
                    p = self.sess.run(var)
                    model_names.append(param_name)
                    params = np.round(p * 10000).astype(np.int).tolist()
                    model_params.append(params)
                    model_shapes.append(p.shape)
        return model_params, model_shapes, model_names

    def set_model_params(self, params):
        with self.g.as_default():
            t_vars = tf.trainable_variables()
            idx = 0
            for var in t_vars:
                if var.name.startswith("MDNTCN"):
                    pshape = tuple(var.get_shape().as_list())
                    p = np.array(params[idx])
                    assert pshape == p.shape, "inconsistent shape"
                    assign_op, pl = self.assign_ops[var]
                    self.sess.run(assign_op, feed_dict={pl.name: p / 10000.0})
                    idx += 1

    def get_random_model_params(self, stdev=0.5):
        # get random params.
        _, mshape, _ = self.get_model_params()
        rparam = []
        for s in mshape:
            # rparam.append(np.random.randn(*s)*stdev)
            rparam.append(np.random.standard_cauchy(s) * stdev)  # spice things up!
        return rparam

    def set_random_params(self, stdev=0.5):
        rparam = self.get_random_model_params(stdev)
        self.set_model_params(rparam)

    def load_json(self, jsonfile="tcn.json"):
        with open(jsonfile, "r") as f:
            params = json.load(f)
        self.set_model_params(params)

    def save_json(self, jsonfile="tcn.json"):
        model_params, model_shapes, model_names = self.get_model_params()
        qparams = []
        for p in model_params:
            qparams.append(p)
        with open(jsonfile, "wt") as outfile:
            json.dump(
                qparams, outfile, sort_keys=True, indent=0, separators=(",", ": ")
            )
        with open(jsonfile.strip(".json")+".hyperparams.pckl", "wb") as f:
            pickle.dump(self.hps, f)
        print("{} written.".format(jsonfile))


def get_pi_idx(x, pdf):
    # samples from a categorial distribution
    N = pdf.size
    accumulate = 0
    for i in range(0, N):
        accumulate += pdf[i]
        if accumulate >= x:
            return i
    random_value = np.random.randint(N)
    # print('error with sampling ensemble, returning random', random_value)
    return random_value


def sigmoid(x):
    return 1 / (1 + np.exp(-x))



from gym import spaces
import numpy as np
import os

from navrep.envs.markenv import MarkEnv, DEFAULT_MAPS
from navrep.envs.encodedenv import EnvEncoder

class MarkOneEncoder(EnvEncoder):
    def __init__(self, backend, encoding,
                 gpu=False):
        assert backend == "VAE_LSTM"
        super(MarkOneEncoder, self).__init__(
            backend, encoding,
            rnn_model_path=os.path.expanduser("~/navrep/models/M/markonernn.json"),
            vae_model_path=os.path.expanduser("~/navrep/models/V/markonevae.json"),
            gpu=gpu,
        )

class MarkTwoEncoder(EnvEncoder):
    def __init__(self, backend, encoding,
                 gpu=False):
        assert backend == "VAE_LSTM"
        super(MarkTwoEncoder, self).__init__(
            backend, encoding,
            rnn_model_path=os.path.expanduser("~/navrep/models/M/marktwornn.json"),
            vae_model_path=os.path.expanduser("~/navrep/models/V/marktwovae.json"),
            gpu=gpu,
        )

class MarkOneEncodedEnv(MarkEnv):
    """ takes a (2) action as input
    outputs encoded obs (546) """
    def __init__(self, backend="VAE_LSTM", encoding="V_ONLY",
                 silent=False, maps=DEFAULT_MAPS, max_episode_length=1000, collect_trajectories=True):
        self.encoder = MarkOneEncoder(backend, encoding)
        super(MarkOneEncodedEnv, self).__init__(
            silent=silent, maps=maps, max_episode_length=max_episode_length,
            collect_trajectories=collect_trajectories)
        self.action_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        self.observation_space = self.encoder.observation_space

    def step(self, action):
        action = np.array([action[0], action[1], 0.])  # no rotation
        obs, reward, done, info = super(MarkOneEncodedEnv, self).step(action)
        h = self.encoder._encode_obs(obs, action)
        return h, reward, done, info

    def reset(self, *args, **kwargs):
        self.encoder.reset()
        obs = super(MarkOneEncodedEnv, self).reset(*args, **kwargs)
        h = self.encoder._encode_obs(obs, np.array([0,0,0]))
        return h

    def render(self, mode="human", close=False):
        decoded_scan = self.encoder._get_last_decoded_scan()
        super(MarkOneEncodedEnv, self).render(mode=mode, close=close, lidar_scan_override=decoded_scan)

class MarkTwoEncodedEnv(MarkEnv):
    """ takes a (2) action as input
    outputs encoded obs (546) """
    def __init__(self, backend="VAE_LSTM", encoding="V_ONLY",
                 silent=False, maps=DEFAULT_MAPS, max_episode_length=1000, collect_trajectories=True):
        self.encoder = MarkTwoEncoder(backend, encoding)
        super(MarkTwoEncodedEnv, self).__init__(
            silent=silent, maps=maps, max_episode_length=max_episode_length,
            collect_trajectories=collect_trajectories)
        self.action_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        self.observation_space = self.encoder.observation_space

    def step(self, action):
        action = np.array([action[0], action[1], 0.])  # no rotation
        obs, reward, done, info = super(MarkTwoEncodedEnv, self).step(action)
        h = self.encoder._encode_obs(obs, action)
        return h, reward, done, info

    def reset(self, *args, **kwargs):
        self.encoder.reset()
        obs = super(MarkTwoEncodedEnv, self).reset(*args, **kwargs)
        h = self.encoder._encode_obs(obs, np.array([0,0,0]))
        return h

    def render(self, mode="human", close=False):
        decoded_scan = self.encoder._get_last_decoded_scan()
        super(MarkTwoEncodedEnv, self).render(mode=mode, close=close, lidar_scan_override=decoded_scan)



from datetime import datetime
import os

from stable_baselines import PPO2
from stable_baselines.common.policies import MlpPolicy
from stable_baselines.common.vec_env import DummyVecEnv

from navrep.envs.markencodedenv import MarkTwoEncodedEnv
from navrep.envs.markenv import SECOND_TRAIN_MAPS
from navrep.tools.sb_eval_callback import NavrepLogCallback
from navrep.tools.commonargs import parse_common_args

args, _ = parse_common_args()

DIR = os.path.expanduser("~/navrep/models/gym")
LOGDIR = os.path.expanduser("~/navrep/logs/gym")
START_TIME = datetime.now().strftime("%Y_%m_%d__%H_%M_%S")
LOGNAME = "marktwoencodedenv_" + START_TIME + "_PPO"
LOGPATH = os.path.join(LOGDIR, LOGNAME + ".csv")
MODELPATH = os.path.join(DIR, LOGNAME + "_ckpt")
MODELPATH2 = os.path.join(DIR, "marktwoencodedenv_latest_PPO_ckpt")
if not os.path.exists(DIR):
    os.makedirs(DIR)
if not os.path.exists(LOGDIR):
    os.makedirs(LOGDIR)

N_ENVS = 6
env = DummyVecEnv([lambda: MarkTwoEncodedEnv(silent=True, maps=SECOND_TRAIN_MAPS)]*N_ENVS)
cb = NavrepLogCallback(logpath=LOGPATH, savepath=MODELPATH, verbose=1)
model = PPO2(MlpPolicy, env, verbose=1)
model.learn(total_timesteps=40000000, callback=cb)
obs = env.reset()

model.save(MODELPATH)
model.save(MODELPATH2)
print("Model '{}' saved".format(MODELPATH))

del model

model = PPO2.load(MODELPATH)

env = MarkTwoEncodedEnv(silent=True, maps=SECOND_TRAIN_MAPS)
obs = env.reset()
for i in range(512):
    action, _states = model.predict(obs, deterministic=True)
    obs, rewards, done, info = env.step(action)
    if done:
        env.reset()
    env.render()
